/*

  This is a part of LSDynamic.dll source code.
  
  Copyright (C) 2001-2002 Sylvain Rouquette
  rouque_s@epita.fr
  http://www.epita.fr:8000/~rouque_s

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
#include "LSDynamic.h"

CDyn::CDyn(e_dyn e, FILE *f, char *n):
angle(0), r_dest(-1), acc(100), r_acc(100), c_acc(100), a_acc(100)
{
	cornerRelative = false;
	alpha.current = 100;
	alpha.enable = false;
	alpha.initial = alpha.final = 255;
	pos.enable = false;
	dest.enable = false;
	center.enable = false;
	c_dest.enable = false;
	size.enable = false;
	info.type = e;
	load(f, n);
}

/************************************************************************************/

CDyn::CDyn(const class CDyn &dyn)
{
	info		= dyn.info;

	size		= dyn.size;
	pos			= dyn.pos;
	dest		= dyn.dest;
	center		= dyn.center;
	c_dest		= dyn.c_dest;
	r_dest		= dyn.r_dest;
	angle		= dyn.angle;
	acc			= dyn.acc;
	r_acc		= dyn.r_acc;
	c_acc		= dyn.c_acc;
	a_acc		= dyn.a_acc;
	alpha		= dyn.alpha;

	delta_time	= dyn.delta_time;
	cornerRelative = dyn.cornerRelative;
}

/************************************************************************************/

CDyn& CDyn::operator=(const CDyn& dyn)
{
	info		= dyn.info;

	size		= dyn.size;
	pos			= dyn.pos;
	dest		= dyn.dest;
	center		= dyn.center;
	c_dest		= dyn.c_dest;
	r_dest		= dyn.r_dest;
	angle		= dyn.angle;
	acc			= dyn.acc;
	r_acc		= dyn.r_acc;
	c_acc		= dyn.c_acc;
	a_acc		= dyn.a_acc;
	alpha		= dyn.alpha;

	delta_time	= dyn.delta_time;
	cornerRelative = dyn.cornerRelative;

	return(*this);
}

/************************************************************************************/

int ParseCoordinateEx(const string &aString, int defaultVal, int maxVal, bool &rel)
{
	string strVal = strlwr((char*) aString.c_str());

	if(strVal.empty())
		return defaultVal;

	char sign = 0;
	int value = 0;
	int i = 0;

	if (strVal.find_first_of("+-~") == 0)
	{
		sign = strVal[i++];
	}

	while (strVal[i] >= '0' && strVal[i] <= '9')
	{
		value = (value * 10) + (strVal[i++] - '0');
	}

	if (sign == '-' || sign == '~')
	{
		value = -value;
	}

	if (strVal[i] == '%')
	{
		i++;
		value = value * maxVal / 100;
	}

	if (strVal.find('c', i) != string::npos)
	{
		value += maxVal / 2;
	}

	rel = strVal.find('r', i) != string::npos;

	while ( (i = strVal.find_first_of("+-", i)) != string::npos)
	{
		value += atoi(strVal.substr(i++).c_str());
	}

	if (!rel && sign != '~' && (value < 0 || (value == 0 && sign == '-') ))
	{
		value += maxVal;
	}

	return value;
}

int ParseDimensionEx(const string &aString, int defaultVal, int maxVal)
{
	string strVal = strlwr((char*) aString.c_str());

	if(strVal.empty())
		return defaultVal;

	char sign = 0;
	int value = 0;
	unsigned int i = 0;

	if (strVal.find_first_of("+-") == 0)
	{
		sign = strVal[i++];
	}

	while (strVal[i] >= '0' && strVal[i] <= '9')
	{
		value = (value * 10) + (strVal[i++] - '0');
	}

	if (sign == '-')
	{
		value = -value;
	}

	if (strVal[i] == '%')
	{
		i++;
		value = value * maxVal / 100;
	}

	while ( (i = strVal.find_first_of("+-", i)) != string::npos)
	{
		value += atoi(strVal.substr(i++).c_str());
	}

	if (value < 0 || (value == 0 && sign == '-'))
	{
		value += maxVal;
	}

	return max(value, 0);
}

void CDyn::getSize(t_pos *p, const string &token1, const string &token2)
{
	p->x = ParseDimensionEx(p->x_str = token1, p->x, SCREEN_WIDTH);
	p->y = ParseDimensionEx(p->y_str = token2, p->y, SCREEN_HEIGHT);
}
void CDyn::getPos(t_pos *p, const string &token1, const string &token2)
{
	p->x = ParseCoordinateEx(p->x_str = token1, p->x, SCREEN_WIDTH, p->x_relative);
	p->y = ParseCoordinateEx(p->y_str = token2, p->y, SCREEN_HEIGHT, p->y_relative);

	if (!p->x_relative)
		p->x += SCREEN_LEFT;

	if (!p->y_relative)
		p->y += SCREEN_TOP;
}
/*void CDyn::getPos(t_pos *p, char *token1, char *token2)
{
	POINT pt;

	bool bTrueNegX = (*token1 == '~');
	bool bTrueNegY = (*token2 == '~');

	if (bTrueNegX) token1[0] = '-';
	if (bTrueNegY) token2[0] = '-';

	p->x_relative = strchr(strlwr(token1), 'r');
	p->y_relative = strchr(strlwr(token2), 'r');

	pt.x = atoi(token1);
	pt.y = atoi(token2);

	if (!p->x_relative)
	{
		if (bTrueNegX)
			pt.x = SCREEN_LEFT + pt.x;
		else
			pt.x = CONVERT_COORDINATE_X(pt.x);
	}
	if (!p->y_relative)
	{
		if (bTrueNegY)
			pt.y = SCREEN_TOP + pt.y;
		else
			pt.y = CONVERT_COORDINATE_Y(pt.y);
	}

	p->x = pt.x;
	p->y = pt.y;
}*/

/************************************************************************************/

void CDyn::load(FILE *f, char *n)
{
	char configLine[MAX_LINE_LENGTH],
		 token1[MAX_LINE_LENGTH],
		 token2[MAX_LINE_LENGTH],
		 extra[MAX_LINE_LENGTH];
	char *tokens[] = {token1, token2};

	info.name = string(n);
	while(LCReadNextConfig(f, "*dyn", configLine, MAX_LINE_LENGTH))
	{
		(void) LCTokenize (configLine, tokens, 2, extra);
		if (strcmp(token2, (info.type == dyn_linear) ? "~linear" : "~circular") == 0)
			break;

		setup(token2, extra);
	}
}

bool CDyn::setup(char *name, char *value)
{
	char token1[MAX_LINE_LENGTH],
		 token2[MAX_LINE_LENGTH],
		 extra[MAX_LINE_LENGTH];
	char *tokens[] = {token1, token2};

	if (stricmp(name, "hwnd") == 0)
		info.getHWND(trimQuotes(value));
	else if (stricmp(name, "caption") == 0)
		info.getCaption(value);
	else if (stricmp(name, "angle") == 0)
		angle = atoi(value);
	else if (stricmp(name, "flags") == 0)
		cornerRelative = strchr(value, 'c') || strchr(value, 'C');
	else if (stricmp(name, "r-dest") == 0)
		r_dest = atoi(value);
	else if (stricmp(name, "acc") == 0)
		acc = atoi(value);
	else if (stricmp(name, "r-acc") == 0)
		r_acc = atoi(value);
	else if (stricmp(name, "c-acc") == 0)
		c_acc = atoi(value);
	else if (stricmp(name, "a-acc") == 0)
		a_acc = atoi(value);
	else if (stricmp(name, "alpha-acc") == 0)
		alpha.current = atoi(value);
	else
	{
		if (LCTokenize (value, tokens, 2, extra) < 2 && stricmp(token1, "disable") != 0)
			return false;

		if (stricmp(name, "range") == 0)
		{
			info.start_time = atoi(token1);
			info.end_time = atoi(token2);
		}
		else if (stricmp(name, "alpha") == 0)
		{
			if (alpha.enable = stricmp(token1, "disable"))
			{
				alpha.initial = atoi(token1);
				alpha.final = atoi(token2);
			}
		}
		else if (stricmp(name, "size") == 0)
		{
			if (size.enable = stricmp(token1, "disable"))
				getSize(&size, token1, token2);
		}
		else if (stricmp(name, "pos") == 0)
		{
			if (pos.enable = stricmp(token1, "disable"))
				getPos(&pos, token1, token2);
		}
		else if (stricmp(name, "dest") == 0)
		{
			if (dest.enable = stricmp(token1, "disable"))
				getPos(&dest, token1, token2);
		}
		else if (stricmp(name, "center") == 0)
		{
			if (center.enable = stricmp(token1, "disable"))
				getPos(&center, token1, token2);
		}
		else if (stricmp(name, "c-dest") == 0)
		{
			if (c_dest.enable = stricmp(token1, "disable"))
				getPos(&c_dest, token1, token2);
		}
		else
			return false;
	}

	return true;
}

/************************************************************************************/

void CDyn::displayUpdate()
{
	if (size.enable)
		getSize(&size, size.x_str, size.y_str);
	if (pos.enable)
		getPos(&pos, pos.x_str, pos.y_str);
	if (dest.enable)
		getPos(&dest, dest.x_str, dest.y_str);
	if (center.enable)
		getPos(&center, center.x_str, center.y_str);
	if (c_dest.enable)
		getPos(&c_dest, c_dest.x_str, c_dest.y_str);
}

/************************************************************************************/

void CDyn::init(CDynMacro *owner, CDynType *window)
{
	if (!info.has_hwnd)
	{
		if (window->has_hwnd)
			info.getHWND(window->hwnd.c_str());
		else if (owner->info.has_hwnd)
			info.getHWND(owner->info.hwnd.c_str());
		else
			info.getHWND(owner->args.hwnd.c_str());
	}

	if (!info.has_caption)
	{
		if (window->has_caption)
			info.copyCaption(window->caption);
		else if (owner->info.has_caption)
			info.copyCaption(owner->info.caption);
		else
			info.copyCaption(owner->args.caption);
	}

	if (!(info.start_time && info.end_time))
	{
		info.start_time = window->start_time;
		info.end_time = window->end_time;
	}
	else
	{
		info.start_time += window->start_time;
		info.end_time += window->start_time;
	}
	info.cur_time = info.start_time;
	delta_time = (info.end_time - info.start_time) / TimerUpdateDelay;

	alpha.enable = DynStorage->has_alpha;
	info.is_running = 0;
}

/************************************************************************************/

void CDyn::update(long t, bool update_pos, bool reverse)
{
	float d_delta;
	CaptionListIterator it;

	info.cur_time = t;

	if (info.cur_time < info.start_time)
		return;

	if (info.is_running == 0)
	{
		info.is_running = 1;
		if (info.hwnd.compare("LsBoxClass") != 0)
		{
			for (int i = info.caption.size(); i > 0; i--)
				DynStorage->getWindowStar(&info);
		}

		it = info.caption.begin();
		while (it != info.caption.end())
		{
			if ((*it)->init(this))
			{
				delete (*it);
				it = info.caption.erase(it);
			}
			else
				it++;
		}
	}

	if (info.cur_time >= info.end_time)
	{
		info.cur_time = info.end_time;
		info.is_running = 2;
	}

	cur_delta_time	= (info.cur_time - info.start_time) / TimerUpdateDelay;
	d_delta			= (delta_time - cur_delta_time) / delta_time;
	if (acc != 100)
		linear_acc	= (cur_delta_time * (100.0f + d_delta * (100 - acc))) / 100.0f;
	else
		linear_acc = cur_delta_time;

	if (r_acc != 100)
		ray_acc		= (cur_delta_time * (100.0f + d_delta * (100 - r_acc))) / 100.0f;
	else
		ray_acc = cur_delta_time;

	if (c_acc != 100)
		center_acc	= (cur_delta_time * (100.0f + d_delta * (100 - c_acc))) / 100.0f;
	else
		center_acc = cur_delta_time;

	if (a_acc != 100)
		angle_acc	= (cur_delta_time * (100.0f + d_delta * (100 - a_acc))) / 100.0f;
	else
		angle_acc = cur_delta_time;

	if (alpha.current != 100)
		alpha.delta	= (cur_delta_time * (100.0f + d_delta * (100 - alpha.current))) / 100.0f;
	else
		alpha.delta = cur_delta_time;

	if (reverse)
	{
		CaptionList::reverse_iterator rit;
		for (rit = info.caption.rbegin(); rit != info.caption.rend(); rit++)
			(*rit)->update(this, update_pos);
	}
	else
		for (it = info.caption.begin(); it != info.caption.end(); it++)
			(*it)->update(this, update_pos);
}

/************************************************************************************/

CDyn::~CDyn()
{
}
