/*

  This is a part of LSDynamic.dll source code.
  
  Copyright (C) 2001-2002 Sylvain Rouquette
  rouque_s@epita.fr
  http://www.epita.fr:8000/~rouque_s

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
#include <windows.h>
#include <time.h>
#include "LSDynamic.h"

using namespace std;

const char RCS_REVISION[] = "$Revision: 1.61 (Syl) $"; // Our Version 
const char RCS_ID[] = "$Id: LSDynamic - ver. 1.61 (Syl) Exp $"; // The Full RCS ID.

char szAppName[] = "LSDynamicModule";
CDynStorage* DynStorage;

int TimerUpdateDelay = 0;

char *trimQuotes(char *buf)
{
	if (StrLen(buf) > 1)
	{
		if ( buf[0] == '"' )
			memmove(buf, buf+1, StrLen(buf));
		if ( buf[StrLen(buf)-1] == '"' )
			buf[StrLen(buf)-1] = '\0';
	}

	return buf;
}

/************************************************************************************/

bool getArgs(StringVector &sv, const char *arguments)
{
	_TCHAR command[MAX_LINE_LENGTH];
	LPCTSTR nextCommand = arguments;

	if (!StrLen(arguments))
		return (false);

	while (GetToken(nextCommand, command, &nextCommand, true))
		sv.push_back(string(command));

	return (true);
}

/************************************************************************************/

CDynStorage::CDynStorage(HWND parentWnd, int& code):
Window(szAppName),
is_running(false), b_bang(false), b_timer(false)
{
	char configLine[MAX_LINE_LENGTH],
		 token1[MAX_LINE_LENGTH],
		 extra[MAX_LINE_LENGTH];
	FILE *f;
	char *tokens[] = {token1};
	bool load_foired = true;
	LPCTSTR ptr1;
	_TCHAR ptr2[MAX_LINE_LENGTH];

	is_running = false;
	if (!createWindow(WS_EX_TOOLWINDOW, szAppName, WS_POPUP, 0, 0, 0, 0, parentWnd))
	{
		code = 1;
		return;
	}

	TimerUpdateDelay = 10;
	OSVERSIONINFO os = { sizeof(os) };
	GetVersionEx(&os);
	has_alpha = ( VER_PLATFORM_WIN32_NT == os.dwPlatformId && os.dwMajorVersion >= 5 );

	if (has_alpha)
	{
		HMODULE hUser32 = GetModuleHandle(_T("USER32.DLL"));
		SetLayeredWindowAttributes = (lpfnSetLayeredWindowAttributes)GetProcAddress(hUser32, "SetLayeredWindowAttributes");
		has_alpha = SetLayeredWindowAttributes != NULL;
	}

	GetRCString("DynUpdateTimer", configLine, "10", MAX_LINE_LENGTH);
	TimerUpdateDelay = atoi(configLine);
	storeBadHandle = GetRCBool("DynStoreBadHandle", true);
	storeHandle = GetRCBool("DynNoHandleStorage", false);
	restoreLayer = GetRCBool("DynForceLayerRestoration", true);
	b_layeredShortcut = GetRCBool("DynLayeredShortcut", true);
	relativeWindowCenter = GetRCBool("DynNoCircularRelativeToWindowCenter", false);
	GetRCString("DynInternalAlpha", configLine, "", MAX_LINE_LENGTH);
	ptr1 = configLine;
	while (GetToken(ptr1, ptr2, (const char **)&ptr1, true))
		internalAlpha.push_back(string(ptr2));

	AddBangCommandEx("!DynModuleHook", LsboxModuleHook);
	AddBangCommandEx("!DynRunning", DynRunning);
	AddBangCommandEx("!DynChangeMovement", DynChangeMovement);
	AddBangCommandEx("!DynPauseAll", DynPauseAll);
	AddBangCommandEx("!DynResumeAll", DynResumeAll);

	f = LCOpen(NULL);
	if (f)
	{
		while(LCReadNextConfig(f, "*DynFile", configLine, MAX_LINE_LENGTH))
		{
			(void) LCTokenize (configLine, tokens, 1, extra);
			createDyns(trimQuotes(extra));
			load_foired = false;
		}
		LCClose(f);
	}
	if (load_foired)
		createDyns();
	code = 0;
}

/************************************************************************************/

CDynStorage::~CDynStorage()
{
	destroyWindow();
}

/************************************************************************************/

void CDynStorage::addDynMacro(FILE *f, char *n)
{
	CDynMacro *dyn = new CDynMacro(f, n);
	dynMacroVector.push_back(dyn);
}

/************************************************************************************/

void CDynStorage::addDyn(FILE *f, e_dyn e, char *n)
{
	CDyn *dyn = new CDyn(e, f, n);
	dynVector.push_back(dyn);
}

/************************************************************************************/

void CDynStorage::createDyns(char *filename)
{
	char configLine[MAX_LINE_LENGTH],
		 token1[MAX_LINE_LENGTH],
		 token2[MAX_LINE_LENGTH],
		 extra[MAX_LINE_LENGTH];
	FILE *f;
	char *tokens[] = {token1, token2};

	f = LCOpen(filename);
	if (!f)
		return;

	while(LCReadNextConfig(f, "*Dyn", configLine, MAX_LINE_LENGTH))
	{
		(void) LCTokenize (configLine, tokens, 2, extra);
		if (stricmp(token2, "macro") == 0)
			addDynMacro(f, trimQuotes(extra));
		else if (stricmp(token2, "linear") == 0)
			addDyn(f, dyn_linear, trimQuotes(extra));
		else if (stricmp(token2, "circular") == 0)
			addDyn(f, dyn_circular, trimQuotes(extra));
	}
	LCClose(f);
}

/************************************************************************************/

bool CDynStorage::hasInternalAlpha(string &str)
{
	for (StringVector::iterator it = internalAlpha.begin(); it != internalAlpha.end(); it++)
		if ((*it).compare(str) == 0)
			return true;
	return false;
}

/************************************************************************************/

void CDynStorage::destroyDyns()
{
	int i;

	is_running = false;
	RemoveBangCommand("!DynModuleHook");
	RemoveBangCommand("!DynRunning");
	RemoveBangCommand("!DynChangeMovement");
	RemoveBangCommand("!DynPauseAll");
	RemoveBangCommand("!DynResumeAll");

	while(b_timer);

	if (!handleWindows.empty())
	{
		for (CaptionListIterator it = handleWindows.begin(); it != handleWindows.end(); it++)
		{
			(*it)->restoreLayer();
			delete (*it);
		}
		handleWindows.clear();
	}

	for(CDynMacroListIterator it = dynMacroStack.begin(); it != dynMacroStack.end(); it++)
	{
		KillTimer(hWnd, (*it)->args.is_running);
		delete *it;
	}
	dynMacroStack.clear();

	for (i = 0; i < dynMacroVector.size(); i++)
	{
		RemoveBangCommand(dynMacroVector[i]->info.name.c_str());
		delete dynMacroVector[i];
	}
	dynMacroVector.clear();

	for (i = 0; i < dynVector.size(); i++)
		delete dynVector[i];
	dynVector.clear();

}

/************************************************************************************/

void CDynStorage::windowProc(Message& message)
{
  BEGIN_MESSAGEPROC
    MESSAGE(onCreate,	WM_CREATE)
    MESSAGE(onDestroy,	WM_DESTROY)
    MESSAGE(onTimer,	WM_TIMER)
    MESSAGE(onDisplayChange,	WM_DISPLAYCHANGE)
    MESSAGE(onGetRevId,	LM_GETREVID)
    MESSAGE(onRefresh,	LM_REFRESH)
  END_MESSAGEPROC
}

/************************************************************************************/

void CDynStorage::onCreate(Message& message)
{
  int lsMessages[] = {LM_GETREVID, LM_REFRESH, 0};

  SendMessage(hParent, LM_REGISTERMESSAGE, WPARAM(hWnd), LPARAM(lsMessages));

	SetWindowLong(hWnd, GWL_USERDATA, magicDWord);
}

/************************************************************************************/

void CDynStorage::onDestroy(Message& message)
{
  int lsMessages[] = {LM_GETREVID, LM_REFRESH, 0};

	destroyDyns();

  SendMessage(hParent, LM_UNREGISTERMESSAGE, WPARAM(hWnd), LPARAM(lsMessages));
}

/************************************************************************************/

void CDynStorage::onGetRevId(Message& message)
{
  char* buf = (char*)(message.lParam);

  switch (message.wParam)
  {
    case 0:
      sprintf(buf, "LSDynamic.dll: %s", &RCS_REVISION[11]);
      buf[StrLen(buf) - 1] = '\0';
      break;
    case 1:
      StrCopy(buf, &RCS_ID[1]);
      buf[StrLen(buf) - 1] = '\0';
      break;
    default:
      buf[0] = 0;
  }
  message.lResult = StrLen(buf);
}

/************************************************************************************/

void CDynStorage::onRefresh(Message& message)
{
	destroyDyns();
	createDyns();
}

/************************************************************************************/

void CDynStorage::exec(const char *str)
{
	char command[MAX_LINE_LENGTH];
	LPCTSTR args = str;

	if (!StrLen(str))
		return;

	GetToken(args, command, &args, false);

	if(*command == '!')
	{
		ParseBangCommand(hWnd, command, args);
	}
	else
		ShellExecute(hWnd, "open", command, args, NULL, SW_SHOWDEFAULT);
}

/************************************************************************************/

void CDynStorage::pause(const string &s)
{
	for (CDynMacroListIterator it = dynMacroStack.begin(); it != dynMacroStack.end(); it++)
	{
		if ((*it)->info.name.compare(s) == 0)
			(*it)->is_paused = true;
	}
}

/************************************************************************************/

void CDynStorage::resume(const string &s)
{
	for (CDynMacroListIterator it = dynMacroStack.begin(); it != dynMacroStack.end(); it++)
	{
		if ((*it)->info.name.compare(s) == 0)
			(*it)->is_paused = false;
	}
}

/************************************************************************************/

void CDynStorage::onTimer(Message& message)
{
	CDynMacroListIterator it = dynMacroStack.begin();
	CDynMacro *curDynMacro;

	while (b_bang);
	b_timer = true;

	while (it != dynMacroStack.end())
	{
		curDynMacro = *it;
		if (curDynMacro->update())
		{
			it = dynMacroStack.erase(it);
			KillTimer(hWnd, curDynMacro->args.is_running);
			delete (curDynMacro);
		}
		else
			it++;
	}
	if (dynMacroStack.empty())
		is_running = false;

	b_timer = false;
}

/************************************************************************************/

void CDynStorage::onDisplayChange(Message& message)
{
	for (int i = 0; i < dynVector.size(); i++)
		dynVector[i]->displayUpdate();
}

/************************************************************************************/

void LsboxModuleHook(HWND caller, const char *bangCommandName, const char *arguments)
{
	char *handle = strrchr(arguments,' ');
	if (!handle) 
		return;
	*handle = 0;

	HWND hWnd = (HWND)atoi(handle+1);

	char *handle2 = strrchr(arguments,' ');
	if (handle2)
		handle2 = strrchr(handle2,'\\');
	else
		handle2 = strrchr(arguments,'\\');
	handle2++;

	for (CaptionListIterator it = DynStorage->handleWindows.begin(); it != DynStorage->handleWindows.end(); it++)
		if ((*it)->name.compare(handle2) == 0)
		{
			(*it)->hWnd = hWnd;
			break;
		}

	if (it == DynStorage->handleWindows.end())
	{
		CCaption *c = new CCaption();
		c->hwnd = string("LsBoxClass");
		c->hWnd = hWnd;
		c->name = string(handle2);
		DynStorage->handleWindows.push_back(c);
	}	
}

/************************************************************************************/

BOOL CALLBACK CDynStorage::enumWindowsProc(HWND hItWnd, LPARAM lParam)
{
	CCaption *caption = NULL;
	char buffer[MAX_LINE_LENGTH];

	if (lParam != NULL)
		caption = (CCaption*)lParam;
	else
		return FALSE;

	if (GetWindow(hItWnd, GW_CHILD))
		EnumChildWindows(hItWnd, CDynStorage::enumWindowsProc, lParam);
	GetClassName(hItWnd, buffer, MAX_LINE_LENGTH);
	if (caption->hwnd.compare(buffer) == 0)
	{
		if (caption->name.size())
		{
			if (GetWindowText(hItWnd, buffer, MAX_LINE_LENGTH) &&
				(caption->name.compare(buffer) == 0))
			{
				caption->hWnd = hItWnd;
				return FALSE;
			}
		}
		else
		{
			caption->hWnd = hItWnd;
			return FALSE;
		}

	}

	return TRUE;
}

void CDynStorage::getWindow(CCaption *cap)
{
	cap->hWnd = NULL;

	for (CaptionListIterator it = handleWindows.begin(); it != handleWindows.end(); it++)
		if ((cap->hwnd.compare((*it)->hwnd) == 0) &&
			(cap->name.compare((*it)->name) == 0))
		{
			cap->hWnd = (*it)->hWnd;
			break;
		}

	if (!cap->hWnd && (it == handleWindows.end()))
	{
		EnumWindows(CDynStorage::enumWindowsProc, (LPARAM)cap);
		if (storeHandle && (cap->hWnd || storeBadHandle))
		{
			CCaption *c = new CCaption(*cap);
			handleWindows.push_back(c);
		}
	}
}

/************************************************************************************/

BOOL CALLBACK CDynStorage::enumWindowsProcStar(HWND hItWnd, LPARAM lParam)
{
	CDynType *dyn_type = NULL;
	char buffer[MAX_LINE_LENGTH];
	CCaption *caption;
	string *str;
	int i;

	if (lParam != NULL)
		dyn_type = (CDynType*)lParam;
	else
		return FALSE;

	if (GetWindow(hItWnd, GW_CHILD))
		EnumChildWindows(hItWnd, CDynStorage::enumWindowsProcStar, lParam);
	GetClassName(hItWnd, buffer, MAX_LINE_LENGTH);
	if (dyn_type->hwnd.compare(buffer) == 0)
	{
		str = &dyn_type->caption.front()->name;
		GetWindowText(hItWnd, buffer, MAX_LINE_LENGTH);
		i = str->find('*');
		if ((i == 0) ||
			(str->compare(0, i, buffer, 0, i) == 0))
		{
			caption = new CCaption(buffer);
			caption->hWnd = hItWnd;
			dyn_type->caption.push_back(caption);
		}
	}

	return TRUE;
}

void CDynStorage::getWindowStar(CDynType *dyn_type)
{
	/*if (!dyn_type->caption.empty())
	{
		for (CaptionListIterator it = dyn_type->caption.begin();
		it != dyn_type->caption.end(); it++)
			delete (*it);
		dyn_type->caption.clear();
	}*/
	EnumWindows(CDynStorage::enumWindowsProcStar, (LPARAM)dyn_type);
	delete dyn_type->caption.front();
	dyn_type->caption.pop_front();
}

/************************************************************************************/
/************************************************************************************/
/************************************************************************************/

void DynRunning(HWND caller, const char *bangCommandName, const char *arguments)
{
	DynStorage->dynRunning(caller, bangCommandName, arguments);
}

void CDynStorage::dynRunning(HWND caller, const char *bangCommandName, const char *arguments)
{
	string buffer = string("Macros which are currently running:\n\n");

	if (dynMacroStack.empty())
		buffer += string("none");
	else
		for (CDynMacroListIterator it = dynMacroStack.begin(); it != dynMacroStack.end(); it++)
		{
			buffer += (*it)->info.name;
			buffer.append(1, '\n');
		}
	MessageBox(hWnd, buffer.c_str(), szAppName, MB_OK|MB_TOPMOST);
}

/************************************************************************************/

void DynPauseAll(HWND caller, const char *bangCommandName, const char *arguments)
{
	DynStorage->dynPauseAll(caller, bangCommandName, arguments);
}

void CDynStorage::dynPauseAll(HWND caller, const char *bangCommandName, const char *arguments)
{
	for (CDynMacroListIterator it = dynMacroStack.begin(); it != dynMacroStack.end(); it++)
		(*it)->is_paused = true;
}

/************************************************************************************/

void DynResumeAll(HWND caller, const char *bangCommandName, const char *arguments)
{
	DynStorage->dynResumeAll(caller, bangCommandName, arguments);
}

void CDynStorage::dynResumeAll(HWND caller, const char *bangCommandName, const char *arguments)
{
	for (CDynMacroListIterator it = dynMacroStack.begin(); it != dynMacroStack.end(); it++)
		(*it)->is_paused = false;
}

/************************************************************************************/

void DynChangeMovement(HWND caller, const char *bangCommandName, const char *arguments)
{
	DynStorage->dynChangeMovement(caller, bangCommandName, arguments);
}

void CDynStorage::dynChangeMovement(HWND caller, const char *bangCommandName, const char *arguments)
{
	if (!StrLen(arguments))
		return;

	char move_name[MAX_LINE_LENGTH],
		 setting[MAX_LINE_LENGTH],
		 value[MAX_LINE_LENGTH];
	char *tokens[] = { move_name, setting };

	int n = LCTokenize (arguments, tokens, 2, value);

	for (int i = 0; i < dynVector.size(); i++)
	{
		if (dynVector[i]->info.name.compare(move_name) == 0)
		{
			if (stricmp(setting, "type") == 0)
			{
				if (stricmp(value, "linear") == 0)
					dynVector[i]->info.type = dyn_linear;
				if (stricmp(value, "circular") == 0)
					dynVector[i]->info.type = dyn_circular;
			}
			else
				dynVector[i]->setup(setting, value);
		}
	}
}

/*void DynChangeMoveType(HWND caller, const char *bangCommandName, const char *arguments)
{
	DynStorage->dynChangeMoveType(caller, bangCommandName, arguments);
}

void CDynStorage::dynChangeMoveType(HWND caller, const char *bangCommandName, const char *arguments)
{
	char move_name[MAX_LINE_LENGTH];
	LPCTSTR move_type = arguments;

	if (!StrLen(arguments))
		return;

	GetToken(move_type, move_name, &move_type, false);

	for (int i = 0; i < dynVector.size(); i++)
		if (dynVector[i]->info.name.compare(move_name) == 0)
			dynVector[i]->info.type = CDynMacro::retDynType(move_type);
}*/

/************************************************************************************/

void CDynStorage::relay(const char *name)
{
	CDynMacroListIterator it = dynMacroStack.begin();
	CDynMacro *curDynMacro, *bang;
	string args;
	int i;

	while (b_timer);

	while (it != dynMacroStack.end())
	{
		curDynMacro = *it;
		if (curDynMacro->info.name.compare(name) == 0)
		{
			if (curDynMacro->args.has_hwnd)
			{
				args += curDynMacro->args.hwnd;
				for (CaptionListIterator cap = curDynMacro->args.caption.begin();
					 cap != curDynMacro->args.caption.end(); cap++)
						 args += string(" ") + (*cap)->name;
			}
			else
				args += string("default");
			for (i = 0; i < curDynMacro->relay.size(); i++)
			{
				if (*curDynMacro->relay[i].c_str() != '!')
					curDynMacro->relay[i].insert(0, '!');
				bang = dynBangCommand(NULL, curDynMacro->relay[i].c_str(), args.c_str());
			}
			it = dynMacroStack.erase(it);
			delete (curDynMacro);
		}
		else
			it++;
	}

}

/************************************************************************************/

void DynBangCommand(HWND caller, const char *bangCommandName, const char *arguments)
{
	DynStorage->dynBangCommand(caller, bangCommandName, arguments);
}

CDynMacro *CDynStorage::dynBangCommand(HWND caller, const char *bangCommandName, const char *arguments)
{
	CDynMacro *dyn_macro;
	CDyn *dyn;
	int i, j;
	static int count = 0;

	//while (b_timer);
	b_bang = true;

	if (stricmp(arguments, "stop") == 0)
	{
		relay(bangCommandName);
		b_bang = false;
		return (NULL);
	}
	if (stricmp(arguments, "pause") == 0)
	{
		pause(string(bangCommandName));
		b_bang = false;
		return (NULL);
	}
	if (stricmp(arguments, "resume") == 0)
	{
		resume(string(bangCommandName));
		b_bang = false;
		return (NULL);
	}

	for (i = 0; i < dynMacroVector.size(); i++)
	{
		if (dynMacroVector[i]->info.name.compare(bangCommandName) == 0)
		{
			dyn_macro = new CDynMacro(*dynMacroVector[i]);
			dynMacroStack.push_back(dyn_macro);
			dyn_macro->init(arguments, TimerUpdateDelay);
			dyn_macro->args.is_running = ++count;

			for (i = dyn_macro->dynTypeVector.size() - 1; i >= 0; i--)
				for (j = 0; j < dynVector.size(); j++)
				{
					if ((dyn_macro->dynTypeVector[i]->type == dyn_move) &&
						(dyn_macro->dynTypeVector[i]->name == dynVector[j]->info.name))
					{
						dyn = new CDyn(*dynVector[j]);
						dyn_macro->dynVector.push_back(dyn);
						dyn->init(dyn_macro, dyn_macro->dynTypeVector[i]);
						dyn_macro->dynTypeVector.erase(&dyn_macro->dynTypeVector[i]);
					}
				}

			//if (!is_running)
				SetTimer(hWnd, dyn_macro->args.is_running, TimerUpdateDelay, NULL);
			//is_running = true;
			break;
		}
	}

	b_bang = false;
	return(dyn_macro);
}

/************************************************************************************/
/************************************************************************************/
/************************************************************************************/

int initModule(HWND ParentWnd, HINSTANCE dllInst, wharfDataType* wd)
{
	return initModuleEx (ParentWnd, dllInst, wd->lsPath);
}

int initModuleEx(HWND ParentWnd, HINSTANCE dllInst, LPCSTR szPath)
{
  int code;

  Window::init(dllInst);

  DynStorage = new CDynStorage(ParentWnd, code);

  return code;
}

void quitModule(HINSTANCE dllInst)
{
	delete DynStorage;
}
